#!/usr/bin/env python2.7
import argparse
import sys
import datetime
import os
import logging
import tempfile
from functools import partial
from multiprocessing import Pool
import VariantUtil
from H5Reader import H5Reader
from ChunkVariantCaller import multi_chunk_var_call,single_chunk_var_call
from _version import version as __version__

def multi_chunk_run(consensus, input, output, chunksize, numpro, tempdir, reference):

    if (tempdir==None):
        tempdir = tempfile.gettempdir()

    # Setting up logging
    logname = os.path.split(output)[-1].split(".")[0]+ '-' + \
              datetime.datetime.now().strftime('%Y%m%d_%H%M%S')+".log"
    # Provide more extensive feedback on the run if --debug is flagged
    if debug_mode:
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s:%(levelname)s:%(name)s:%(message)s',
                            filename=logname,
                            filemode='a')
    else:
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)s:%(levelname)s:%(name)s:%(message)s',
                            filename=logname,
                            filemode='a')
    logger = logging.getLogger(__name__)

    reader=H5Reader(input,chunksize,20)
    reader.getReadsInterval()
    reader.getChunkAlignmentIDs()
    numchunks = reader.genomelength / reader.intervalLength + 1
    logger.info("Chunking the sequence into " + str(numchunks) + " chunks with length of " + str(reader.intervalLength))

    partial_h5pbhoover=partial(multi_chunk_var_call,input=input,alignmentIDs=reader.alnIDs,
                               tempdir=tempdir,numberOfReads=reader.numberOfReads)
    if(numpro==0):
        pool=Pool()
    else:
        pool=Pool(processes=numpro)
    results=pool.map(partial_h5pbhoover,reader.interval)
    pool.close()
    pool.join()
    # Left-align indels using bcftools norm
    normalized_results = VariantUtil.normalize(results_list=results,
                                               reference=reference,
                                               nproc=numpro)
    # Combine list of lists into a single list of one list and save as results
    merged_results = [] 
    intermed_list = []
    for record_list in normalized_results:
        intermed_list += record_list
    records_to_process, single_records = VariantUtil.set_consecutive_entries(normalized_results)
    intermediate_list = VariantUtil.combine_variants(records_to_process) + single_records
    sorted_records = sorted(intermediate_list, key=lambda x: x.POS)  

    # Sort duplicates and have preference for PASS, then HETERO over NO/LOW
    for i, record in enumerate(sorted_records):
        if i != 0:
            # Process duplicate
            if record.POS == prev_entry.POS:
                # current record should be switched with current first duplicate
                if not record.FILTER:
                    sorted_records[i-1],sorted_records[i] = sorted_records[i], sorted_records[i-1]
                elif record.FILTER and record.FILTER[0] == "HETERO":
                    sorted_records[i-1],sorted_records[i] = sorted_records[i], sorted_records[i-1]
        prev_entry = record
    merged_results.append(sorted_records)
    results = merged_results
    write_to_vcf_multi(output, results, consensus)


def target_region_run(consensus, input, output, interval):

    results=single_chunk_var_call(interval,input)
    write_to_vcf_single(output, results, consensus)


def write_vcf_entry(writer, entry, consensus):
    for vcfline in entry:
        if vcfline == None \
           or (not consensus and vcfline.FILTER[0] in ['NO', 'LOW']):
            continue
        writer.write_record(vcfline)


def write_to_vcf_single(output, results, consensus):

    vcf_writer = VariantUtil.init_vcf(output)
    write_vcf_entry(vcf_writer, results, consensus)
    vcf_writer.close()


def write_to_vcf_multi(output, results, consensus):

    vcf_writer = VariantUtil.init_vcf(output)
    for vcf_entry in results:
        write_vcf_entry(vcf_writer, vcf_entry, consensus)
    vcf_writer.close()

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--version',
                        help='Print program version and exit',
                        action='store_true')
    parser.add_argument('-c',
                        '--consensus',
                        help='Call consensus (output all positions to VCF)',
                        action='store_true')
    parser.add_argument('-i',
                        '--input',
                        help='Input cmp.h5 file')
    parser.add_argument('-o',
                        '--output',
                        help='Output VCF file name, stdout by default',
                        default=sys.stdout)
    parser.add_argument('-r',
                        '--reference',
                        help='Reference FASTA file for indel normalization')
    parser.add_argument('-n',
                        '--nproc',
                        type=int,
                        help='number of processors to be used, uses all available by default',
                        default=0)
    parser.add_argument('-s',
                        '--chunksize',
                        help='Genome chunk size used for parallelizaton, 5000 by default',
                        type=int,
                        default=5000)
    parser.add_argument('-t',
                        '--tempdir',
                        help='Temporary directory for analysis. Current directory is default',
                        default=None)
    parser.add_argument('-d',
                        '--debug',
                        help='Flag for debugging purposes. Writes more information to log',
                        action='store_true')
    args = parser.parse_args()

    if args.version:
        print __version__
        sys.exit(0)
    global debug_mode
    debug_mode = args.debug
    multi_chunk_run(args.consensus,args.input,args.output,args.chunksize,args.nproc,args.tempdir, args.reference)

if __name__ == "__main__":
    main()
