from __future__ import division
import os
import operator
from operator import itemgetter
from math import log
import sys
import logging
import subprocess
from collections import OrderedDict, Counter
import PBHoover_Runner
import vcf
import tempfile
sys.path.append('pbhoover/')
from _version import version as __version__


__program__ = 'pbhoover'


def init_vcf(outfile):
    file_handle = open(outfile, 'w') if outfile != sys.stdout else outfile
    vcf_template = vcf.Reader(os.devnull)
    vcf_template.metadata['fileformat'] = 'VCFv4.0'
    vcf_template.metadata['source'] = [__program__ + 'V' + __version__]

    vcf_template.filters['LOW'] = vcf.parser._Filter('LOW','Position with too low of depth')
    vcf_template.filters['NO'] = vcf.parser._Filter('NO', 'Does not meet criteria to call variant')
    vcf_template.filters['HETERO'] = vcf.parser._Filter( 'HETERO',
                                                         'Enough support to call reference and variant '
                                                         '(mixed population)')

    vcf_template.infos['RSR'] = vcf.parser._Info('RSR',1,'Integer','Reference-supporting reads')
    vcf_template.infos['VSR'] = vcf.parser._Info('VSR',1,'Integer','Variant-supporting reads')
    # TODO - This is better represented in VCF natively
    vcf_template.infos['VF'] = vcf.parser._Info('VF',1,'Float','Variant frequency')
    vcf_template.infos['DP'] = vcf.parser._Info('DP',1,'Integer','Read Depth')

    vcf_template._column_headers = [
        'CHROM',
        'POS',
        'ID',
        'REF',
        'ALT',
        'QUAL',
        'FILTER',
        'INFO'
    ]
    return vcf.Writer(file_handle, vcf_template)


def consensus_quality(qualities):
    qualities = {k:v for k, v in qualities.iteritems() if v != []}
    consensus_quality = {}
    for call,quality in qualities.iteritems():
            opposing_qualities = sum({k:v for k, v in qualities.iteritems() if k != call}.values(),[])
            if(opposing_qualities != []):
                for q in opposing_qualities:
                    product = reduce(operator.mul,[10**(-float(q)/10)])
            #product = reduce(operator.mul,[10**(-q/10) for q in opposing_qualities],1)
            else:
                product = 0

            if(product >= 1):
                consensus_quality[call] = 0
            else:
                consensus_quality[call] = sum(quality) - 10*log(1-product,10)

    return consensus_quality

def is_insertion(call):
    return len(call) == 2

def is_deletion(call):
    return call[0] == '-'

def is_multibase_indel(call):
        return len(call) > 2

def is_single_indel(call):
    return is_deletion(call) or is_insertion(call)

def is_snp(call):
    return len(call) == 1 and not is_deletion(call)


def call_consensus(referencecall, position, calls, qualities,multi_del):
    refbase = referencecall
    var_calls = {call:coverage for call,coverage in
                 calls.iteritems() if call != refbase and coverage != 0}
    var_quals = {call:qual for call,qual in
                 qualities.iteritems() if call != refbase}

    if not var_calls.keys():
        return None

    variant = get_top_variant(variant_coverage=var_calls,
                  qualities=var_quals)
    #print variant

    RSR = calls[refbase] if refbase in calls else 0
    VSR = calls[variant]
    var_freq = VSR/(RSR + VSR)
    CHROM = "1"
    # Python and pysam use zero-based counts.
    POS = position + 1
    ID = None
    ALT = None
    QUAL = None
    FILTER = None
    INFO = None
    FORMAT = None
    sample_indexes = {}


    REF = refbase
    if is_insertion(variant):
        ALT = [refbase + variant[1:]]
    elif is_deletion(variant):
        POS -= 1
        REF = refbase
        ALT = "-"
    else:
        ALT = [variant]

    INFO = dict(RSR=RSR,
            VSR=calls[variant],
            VF=var_freq,
            DP=sum(calls.values()))
    #if is_snp(variant):
    QUAL = "%.2f" % qualities[variant]
    #else:
    #    QUAL = None

    FILTER = eval_variant(variant=variant,
                  RSR=INFO['RSR'],
                  VSR=INFO['VSR'],
                  var_freq=INFO['VF'],multi_del=multi_del)
    vcf_entry = vcf.model._Record(
        CHROM,
        POS,
        ID,
        REF,
        ALT,
        QUAL,
        FILTER,
        INFO,
        FORMAT,
        sample_indexes
    )
    return vcf_entry


def normalize(results_list, reference, nproc):
    logger = logging.getLogger('Normalization')
    logger.info('Left aligning variants with bcftools norm')
    # Normalize Variants #
    unnorm_vcf_all = tempfile.NamedTemporaryFile(suffix='.vcf')
    norm_vcf = tempfile.NamedTemporaryFile(suffix='.vcf')
    norm_wrongheader_vcf = tempfile.NamedTemporaryFile(suffix='.vcf')
    PBHoover_Runner.write_to_vcf_multi(unnorm_vcf_all.name, results_list, True)
    logger.debug('Zipping raw VCF')
    with open(unnorm_vcf_all.name + '.gz', 'w') as f:
        bgzip = subprocess.call(['bgzip', '-c', unnorm_vcf_all.name], stdout=f)
    tabix_out = subprocess.call(['tabix', '-p', 'vcf', unnorm_vcf_all.name + '.gz'])

    bcftools_norm = ['bcftools', 'norm', '--threads', str(nproc),
                     '--check-ref', 'w', '--fasta-ref', reference,
                     '--output-type', 'v',
                     '--output', norm_wrongheader_vcf.name,
                     unnorm_vcf_all.name + '.gz']
    stderr = tempfile.TemporaryFile()
    logger.debug('Calling bcftools norm')
    norm = subprocess.call(bcftools_norm, stderr=stderr)
    stderr.close()

    grep_out = subprocess.call(['grep', '-v', '^##contig', norm_wrongheader_vcf.name], stdout=norm_vcf)
    with open(norm_vcf.name + '.gz', 'w') as f:
        bgzip = subprocess.call(['bgzip', '-c', norm_vcf.name], stdout=f)
    tabix = subprocess.call(['tabix', '-p', 'vcf', norm_vcf.name + '.gz'])

    # Closing tempfiles will also delete the file
    # unnorm_vcf_all.close()
    # norm_output.close()

    normalized_vcf_file = vcf.Reader(filename=norm_vcf.name + '.gz')
    normalized_results = []
    logger.debug('Reading in VCF and creating a list of PyVCF records')
    for record in normalized_vcf_file.fetch('1', 1, 5000000):
        if not record.FILTER:
            record.FILTER = ['PASS']
        vcf_entry = vcf.model._Record('1', record.POS, record.ID, record.REF, record.ALT, record.QUAL, record.FILTER,
                                      record.INFO, record.FORMAT, record.samples)
        normalized_results.append(vcf_entry)
    unnorm_vcf_all.close()
    norm_wrongheader_vcf.close()
    norm_vcf.close()
    logger.debug('Finished loading VCF records')
    return normalized_results


def set_consecutive_entries(record_list):
    """Groups consecutive variants into lists contained within a list."""
    logger = logging.getLogger('FindConsecutiveEntries')
    vcf_out = [] # List of records which we will write to the new vcf file
    consecutives, duplicates = [], []
    block, intermediate_list = [],[]
    distance, last_position = 0, 0
    duplicate = False
    counts = 0
    #print "line 162, VariantUtil.py"
    for record in record_list:
        # "record.FILTER is: ", record.FILTER
        # "length of record.FILTER is: ", len(record.FILTER)
        if "PASS" in record.FILTER or 'HETERO' in record.FILTER:

            # intermediate_list.append(record)
        # elif record.FILTER == "HETERO":
        #     intermediate_list.append(record)
        # elif record.FILTER == "NO" or record.FILTER == "LOW":
        #     vcf_out.append(record)
    # for record in intermediate_list:
            distance = record.POS - last_position
            # Hit duplicate entry, reset block and add block
            if distance == 0: # Duplicate entry
                duplicate = True

            # Hit non-consecutive entry or the first entry of a new block, reset block and add block
            elif distance > 1:
            # First entry in entire file, skip
                if counts == 0:
                    counts += 1
                else:
                    if block:
                        # Only add the block to process if the length is greater than one
                        # Extend consecutives list
                        if (len(block) > 1) and (duplicate is False):
                            consecutives.append(block)
                        # Extend duplicates list
                        elif (len(block) > 1) and (duplicate is True):
                            vcf_out += block
                            duplicate = False
                        # Write out single record to vcf file
                        elif len(block) == 1:
                            vcf_out.append(block[0])
                        block = []
            # Consecutive entries, append to current block
            else:
                distance = 0
            block.append(record)
            last_position = record.POS
        elif 'NO' in record.FILTER or "LOW" in record.FILTER:
            vcf_out.append(record)

    # Last block
    if (len(block) > 1) and (duplicate is False):
        consecutives.append(block)
    elif (len(block) > 1) and (duplicate is True):
        vcf_out += block
    elif len(block) == 1:
        vcf_out.append(block[0])
    elif len(block) == 0:
        pass
    return consecutives, vcf_out # Return list of consec records and list of consecutive records

def check_type(entry):
    logger = logging.getLogger('CheckType')
    ref = entry.REF
    alt = entry.ALT[0]

    if len(ref)==len(alt):
        return "snp", 0
    if len(ref)<len(alt):
        insert_len=len(alt)-len(ref)
        return "insertion", insert_len
    if len(ref)>len(alt):
        delete_len=len(ref)-len(alt)
        return "deletion", delete_len
    else:
        # print("something is wrong")
        logger.warn('Type of variant not found')

def populate_list(ref_alt,pos):
    # List will contain base and position: ex-[['A',301],['C',302]]
    ref_alt_list=[]
    for i,base in enumerate(ref_alt):
        ref_alt_list.append((base,i+pos))
    return ref_alt_list

def contains_del_snp(group):
    index = -1
    prev_type = ""
    for i,item in enumerate(group):
        cur_type = check_type(item)[0]
        if (cur_type == "snp") and (prev_type == "deletion"):
            index = i - 1
            return  index
        prev_type = check_type(item)[0]
    return index

    # Checks if the group of consecutive variants passed in has consecutive snps, returns if there are either consecutive snps or dels, the index of the first occurrence and the length
def contains_consec(group, snp_or_del):
    contains_consec = -1 # Index representing start of consec snps/dels
    length = -1 # Number of consecutive snps/dels counted in total
    prev_type = ""
    snp_counts = 1
    del_counts = 1
    prev_type = "" # simulates parent pointer
    prev_entry, cur_entry = None,None

    for index,cur_entry in enumerate(group):
        cur_type = check_type(cur_entry)[0]

        # First entry in group
        if index == 0:
            prev_type = cur_type
            prev_entry = cur_entry
        else:
        # Found at least 2 consecutive entries that must be snp/snp or del/del
            if (cur_type == snp_or_del) and (prev_type == snp_or_del):
                next_pos = prev_entry.POS + 1
                # Increase counts only if positions are also consecutive
                if cur_entry.POS == next_pos:
                    if prev_type == "snp":
                        snp_counts += 1
                        if contains_consec == -1:
                            # Set start index of consec snps
                            contains_consec = index - 1

                    elif prev_type == "deletion":
                        del_counts += 1
                        if contains_consec == -1:
                            # Set start index of consec dels
                            contains_consec = index - 1

            elif prev_type != cur_type or index == (len(group) - 1):
                if (snp_counts > 1) and (snp_or_del == "snp"):
                    length = snp_counts
                    return contains_consec,length
                elif (del_counts > 1) and (snp_or_del == "deletion"):
                    length = del_counts
                    return contains_consec,length

        # Update parent pointer to last current entry
        prev_entry = cur_entry
        prev_type = check_type(prev_entry)[0]
    if length == -1:
        if snp_counts > 1 and snp_or_del == 'snp':
            length = snp_counts
        elif del_counts > 1 and snp_or_del == 'deletion':
            length = del_counts
    # No consecutive snps or dels found
    return contains_consec,length

# Function checks if the current group of consecutive entries has contains either a ins/del or del/ins
# Input: group, Output: (index of first occurrence of ins/del or del/ins)
def ins_del(group):
    prev_entry_type, cur_entry_type = "",""
    for index,entry in enumerate(group):
        cur_entry_type = check_type(entry)[0]
        if index != 0:
            if (cur_entry_type == "deletion") and (prev_entry_type == "insertion"):
                return index-1,"ins_del"
            elif (cur_entry_type == "insertion") and (prev_entry_type == "deletion"):
                return index-1,"del_ins"
        prev_entry_type = cur_entry_type
    return -1, None # No ins_del or del_ins found

def combine_variants(consecutive_list):
    logger = logging.getLogger('CombineVariants')
    logger.info("Combining variants")
    compare = lambda x, y: Counter(x) == Counter(y)
    merged_variants = []
    previous_id, current_id = "", ""
    skip_entry = False
    prev_entry = ""
    counter_var = 0
    logger.debug('Number of consecutive blocks ' + str(len(consecutive_list)))
    for group in consecutive_list:
        if len(group) > 2:
            group_size = len(group)
            snp_count, ins_count, del_count = 0, 0, 0
            unknown_cases, unknown_refs_alts = [],[]
            prev_id, entry_type = "",""
            snp_counts = 1 # Tracks number of additional times current type of variant occurs (ex: 3 snps in a row)
            del_counts = 1

            for index,entry in enumerate(group):
                print entry.POS, entry.REF, entry.ALT
                entry_type = check_type(entry)[0]
                if entry_type == "snp":
                    snp_count += 1
                    unknown_cases.append("snp")
                elif entry_type == "insertion":
                    ins_count += 1
                    unknown_cases.append("insertion")
                elif entry_type == "deletion":
                    del_count += 1
                    unknown_cases.append("deletion")
                unknown_refs_alts.append([entry.POS, entry.REF, str(entry.ALT[0])])

            if snp_count == group_size:
                logger.debug('SNPs' + ','.join([str(entry.POS) for entry in group]))
                combined_ref, combined_alt  = "",""
                for index,entry in enumerate(group):
                    combined_ref += entry.REF
                    combined_alt += str(entry.ALT[0])

                HETERO = False
                filter = ""
                lowest_qual = sys.maxint
                index = -1 # Position of record in group with lowest quality
                for i,record in enumerate(group):
                    if record.FILTER:
                        if 'HETERO' in record.FILTER:
                            filter = record.FILTER
                            HETERO = True
                        if record.QUAL < lowest_qual:
                            index = i # Store index with lowest quality score
                            lowest_qual = record.QUAL

                if HETERO is False:
                    filter = group[0].FILTER
                # Assign rest of columns from record with lowest quality score using index
                entry = group[index]
                chrom = entry.CHROM
                pos = group[0].POS # Set position to first entry to avoid REF_MISMATCH error
                id = entry.ID
                qual = entry.QUAL
                info = entry.INFO
                format = entry.FORMAT
                samples = entry.samples

                vcf_entry = vcf.model._Record(chrom, pos, id, combined_ref, [combined_alt], qual, filter, info, format,
                    samples)
                merged_variants.append(vcf_entry)
                continue # Go to next group

            elif ins_count > 1:
                logger.debug('Insertions' + ','.join([str(entry.POS) for entry in group]))
                for item in group:
                    merged_variants.append(item)
                continue

            # All deletions: unique union of refs and only keep first alt base
            elif del_count == group_size:
                logger.debug('Deletions' + ','.join([str(entry.POS) for entry in group]))
                # Get ref (unique union of all consecutive refs)
                set_refs = []
                combined_refs = ""
                for entry in group:
                    intermediate = populate_list(entry.REF, entry.POS)
                    set_refs += intermediate
                set_refs = set(set_refs)
                sorted_refs = sorted(set_refs, key=itemgetter(1)) # Sort set of items by position
                for index,entry in enumerate(sorted_refs):
                    combined_refs += entry[0]

                first_entry = group[0] # pyvcf record object
                combined_alts = [first_entry.ALT[0]] # only the first letter in the tuple will be the new alt

                HETERO = False
                lowest_qual = sys.maxint
                index = -1 # Position of record in group with lowest quality
                filter = ""
                for i,record in enumerate(group):
                    if record.FILTER:
                        if 'HETERO' in record.FILTER:
                            filter = record.FILTER
                            HETERO = True
                        if record.QUAL < lowest_qual:
                            index = i
                            lowest_qual = record.QUAL
                if HETERO is False:
                    filter = group[0].FILTER
                entry = group[index]
                chrom = entry.CHROM
                pos = entry.POS
                id = entry.ID
                qual = entry.QUAL
                info = entry.INFO
                format = entry.FORMAT
                samples_indexes = {}

                vcf_entry = vcf.model._Record(chrom, pos, id, combined_refs, combined_alts, qual, filter, info, format,
                                              samples_indexes)
                merged_variants.append(vcf_entry)
                continue

            if contains_consec(group,"snp")[0] != -1:
                logger.debug('Merging all SNPs')
                combined_ref,combined_alt = "",""
                start_position = contains_consec(group,"snp")[0] # Keep original copy of start
                start_pos_copy = start_position # Create copy of start position to modify
                consec_types = contains_consec(group,"snp")[1] # Stores number of variants of the same type in a row

                for x in range(consec_types):
                    # cur_record is a pyvcf record
                    cur_record = group[start_pos_copy]
                    combined_ref += cur_record.REF
                    combined_alt += str(cur_record.ALT[0])
                    start_pos_copy += 1 # Go to next consecutive position

                HETERO = False
                lowest_qual = sys.maxint
                index = -1 # Position of record in group with lowest quality
                filter = ""
                for i,record in enumerate(group):
                    if record.FILTER:
                        if record.FILTER[0] == "HETERO":
                            filter = record.FILTER
                            HETERO = True
                        if record.QUAL < lowest_qual:
                            index = i
                            lowest_qual = record.QUAL

                if HETERO is False:
                    filter = group[0].FILTER
                entry = group[index]
                chrom = entry.CHROM
                pos = group[start_position].POS
                id = entry.ID
                qual = entry.QUAL
                info = entry.INFO
                format = entry.FORMAT
                samples = entry.samples

                new_record = vcf.model._Record(chrom, pos, id, combined_ref, [combined_alt], qual, filter, info, format,
                                               samples)
                group[start_position:start_position + consec_types] = [new_record]

            if contains_consec(group,"deletion")[0] != -1:
                logger.debug('Merging all Deletions')
                combined_ref, combined_alt = "",""
                start_position = contains_consec(group,"deletion")[0]  # Keep original copy of start
                start_pos_copy = start_position # Create copy of start position to modify
                consec_types = contains_consec(group,"deletion")[1] # Stores number of variants of the same type in a row
                first_record = group[start_position]
                combined_refs = set()
                combined_alt = [first_record.ALT[0]]
                refs = [] # will hold unique union of all refs in group

                for x in range(consec_types):
                    # Current record is a pyvcf record
                    cur_record = group[start_pos_copy]
                    cur_ref_list = populate_list(cur_record.REF,cur_record.POS) # Ex: return [['A', 306],['C', 307]] --> [base,position]
                    refs += tuple(cur_ref_list) # Convert current base to a tuple
                    start_pos_copy += 1 # Go to next consecutive position
                combined_refs = sorted(set(refs), key=lambda x: x[1]) # sorted, populated list of all the ref bases, sort by positions
                for item in combined_refs:
                    combined_ref += item[0]

                first_pos = first_record.POS
                HETERO = False
                lowest_qual = sys.maxint
                index = -1  # Position of record in group with lowest quality
                FILTER = ""
                for i, record in enumerate(group):
                    if record.FILTER:
                        if 'HETERO' in record.FILTER:
                            FILTER = record.FILTER
                            HETERO = True
                        if record.QUAL < lowest_qual:
                            index = i
                            lowest_qual = record.QUAL

                if HETERO is False:
                    FILTER = group[0].FILTER
                entry = group[index]
                chrom = entry.CHROM
                pos = first_pos
                ID = entry.ID
                qual = entry.QUAL
                info = entry.INFO
                FORMAT = entry.FORMAT
                samples = entry.samples

                new_record = vcf.model._Record(chrom, pos, ID, combined_ref, combined_alt, qual, FILTER, info, FORMAT,
                                               samples)
                group[start_position:start_position + consec_types] = [new_record]
            prev_type, cur_type = "","" # simulates parent and child points to pyvcf record objects
            #ins_del, del_ins = [],[] # Store indices of start positions from intermediate list

            if ins_del(group)[0] != -1:
                if ins_del(group)[1] == "ins_del":
                    logger.debug('Merging Insertion/Deletion ' + ','.join([str(group[ins_del(group)[0]].POS),
                                                                           str(group[ins_del(group)[0] + 1].POS)]))
                    # unique union for refs
                    index = ins_del(group)[0] # index of the ins
                    prev_entry = group[index]
                    ref1 = populate_list(group[index].REF, group[index].POS) # Corresponds to the insertion
                    ref2 = populate_list(group[index+1].REF, group[index+1].POS) # Corresponds to the deletion
                    alt1 = populate_list(str(group[index].ALT[0]), group[index].POS)
                    alt2 = populate_list(str(group[index+1].ALT[0]), group[index+1].POS)

                    combined_ref = sorted(set(ref1+ref2), key=itemgetter(1))
                    combined_alt = sorted(set(alt1+alt2), key=itemgetter(1))
                    new_refs,new_alts="",""

                    if len(combined_ref) == len(combined_alt): # Iterate through list of: [base,position]
                        first_occurence = -1
                        # Iterate through tuples: ex: ('A', 303)
                        for i,j in zip(combined_ref,combined_alt):
                            if i[0]!=j[0]:
                                # store position of first mismatch, make new reference genome position
                                if first_occurence == -1:
                                    new_pos = i[1]
                                    first_occurence += 1
                                new_refs += i[0]
                                new_alts += j[0]
                    else:
                        # lengths of combined ref and alt are not equal, don't merge entire block
                        for record in group:
                            merged_variants.append(record)
                        break # Do not check for anymore ins/dels

                    # Search for at least one hetero and index with lowest quality score
                    HETERO = False
                    lowest_qual = sys.maxint
                    index = -1 # Position of record in group with lowest quality
                    filter = ""
                    for i,record in enumerate(group):
                        if record.FILTER:
                            if 'HETERO' in record.FILTER:
                                filter = record.FILTER
                                HETERO = True
                            if record.QUAL < lowest_qual:
                                index = i
                                lowest_qual = record.QUAL

                    if HETERO is False:
                        filter = group[0].FILTER
                    entry = group[index]
                    chrom = entry.CHROM
                    pos = new_pos
                    id = entry.ID
                    qual = entry.QUAL
                    info = entry.INFO
                    format = entry.FORMAT
                    samples = entry.samples

                    vcf_entry = vcf.model._Record(chrom, pos, id, new_refs, [new_alts], qual, filter, info, format,
                                                  samples)
                    group[index:index+2] = [vcf_entry]
                    continue # Check for more ins/del or del/ins

                elif ins_del(group)[1] == "del_ins":
                    logger.debug('Merging Deletion/Insertion ' + ','.join([str(group[ins_del(group)[0]].POS),
                                                                           str(group[ins_del(group)[0] + 1].POS)]))
                    index = ins_del(group)[0] # index of deletion
                    prev_entry = group[index]
                    ref1 = populate_list(group[index].REF, group[index].POS)
                    ref2 = populate_list(group[index+1].REF, group[index+1].POS)
                    alt1 = populate_list(str(group[index].ALT[0]), group[index].POS)
                    alt2 = populate_list(str(group[index+1].ALT[0]), group[index+1].POS)
                    combined_ref = sorted(set(ref1+ref2), key=itemgetter(1)) # unique union of populated list
                    combined_alt = alt1 # del's alt populated list

                    for index,entry in enumerate(alt2):
                        if index != 0:
                            combined_alt.append(entry)
                    combined_alt = sorted(set(combined_alt), key=itemgetter(1))

                    if len(combined_ref)==len(combined_alt):
                        new_ref,new_alt,new_pos= "","",""
                        first_occurence = -1
                        for i,j in zip(combined_ref,combined_alt):
                            if i[0]!=j[0]:
                                if first_occurence == -1:
                                    new_pos = i[1]
                                    first_occurence += 1
                                new_ref += i[0]
                                new_alt += j[0]

                        HETERO = False
                        lowest_qual = sys.maxint
                        index = -1 
                        filter = ""
                        for i,record in enumerate(group):
                            if record.FILTER:
                                if 'HETERO' in record.FILTER:
                                    filter = record.FILTER
                                    HETERO = True
                                if record.QUAL < lowest_qual:
                                   index = i
                                   lowest_qual = record.QUAL

                        if HETERO is False:
                            filter = group[0].FILTER
                        entry = group[index]
                        chrom = entry.CHROM
                        pos = new_pos
                        id = entry.ID
                        qual = entry.QUAL
                        info = entry.INFO
                        format = entry.FORMAT
                        samples = entry.samples
                        vcf_entry = vcf.model._Record(chrom, pos, id, new_ref, [new_alt], qual, filter, info, format,
                                                      samples)
                        group[index:index+2] = [vcf_entry]
                    else: # Lengths are not equal --> write out entire block
                        for item in group:
                            merged_variants.append(item)
                        break # do not check for anymore del/ins

            if contains_consec(group,"snp")[0] != -1: # value will be -1 if there is no start index for a set of consecutive snps found
                combined_ref,combined_alt = "",""
                start_position = contains_consec(group,"snp")[0] # Keep original copy of start
                start_pos_copy = start_position # Create copy of start position to modify
                consec_types = contains_consec(group,"snp")[1] # Stores number of variants of the same type in a row
                record_template = group[start_position]
                index = contains_consec(group,"snp")[0]

                chrom = record_template.CHROM
                pos = record_template.POS
                id = record_template.ID
                qual = record_template.QUAL
                filter = record_template.FILTER
                info = record_template.INFO
                format = record_template.FORMAT
                samples = record_template.samples

                for x in range(consec_types): # Simple concatenation
                    # Current record is a pyvcf record, simply concatenate refs and alts since merging snps
                    cur_record = group[start_pos_copy]
                    combined_ref += cur_record.REF
                    combined_alt += str(cur_record.ALT[0])
                    start_pos_copy += 1 # Go to next consecutive position

                HETERO = False
                filter = ""
                lowest_qual = sys.maxint
                index = -1 # Position of record in group with lowest quality
                for i,record in enumerate(group):
                    if record.FILTER:
                        if 'HETERO' in record.FILTER:
                            filter = record.FILTER
                            HETERO = True
                        if record.QUAL < lowest_qual:
                            index = i
                            lowest_qual = record.QUAL

                if HETERO == False:
                       filter = group[0].FILTER
                entry = group[index]
                chrom = entry.CHROM
                pos = new_pos
                id = entry.ID
                qual = entry.QUAL
                info = entry.INFO
                format = entry.FORMAT
                samples = entry.samples
                new_record = vcf.model._Record(chrom, pos, id, combined_ref, [combined_alt], qual, filter, info, format,
                                               samples)
                group[start_position:start_position + consec_types] = [new_record]

            # At this point, all ins/del, del/ins, consecutive snps, and consecutive dels, and are combined into new group
            if len(group) == 3:
                logger.debug('Merging SNP/Insertion/SNP or Insertion/SNP/Deletion' +
                             ','.join([str(e.POS) for e in group]))
                if contains_del_snp(group) != -1:
                    del_index = contains_del_snp(group)
                    del group[del_index]

                elif (compare(unknown_cases, ['snp', 'insertion', 'snp']) == True) or \
                        (compare(unknown_cases, ["insertion", "snp", "deletion"]) == True):
                    concatenated_ref = []
                    concatenated_alt = []
                    for group_record in group:
                        concatenated_ref.append((group_record.REF, group_record.POS))
                        concatenated_alt.append((str(group_record.ALT[0]), group_record.POS))

                    combined_ref = sorted(set(concatenated_ref),key=itemgetter(1))
                    combined_alt = sorted(set(concatenated_alt),key=itemgetter(1))
                    counter = 0
                    new_refs,new_alts = "",""
                    if len(combined_ref) == len(combined_alt):
                        first_occurence = -1
                        for i,j in zip(combined_ref,combined_alt):
                            if i[0] != j[0]:
                                if first_occurence == -1:
                                    new_pos = i[1]
                                    first_occurence += 1
                                new_refs += i[0]
                                new_alts += j[0]

                    HETERO = False
                    filter = ""
                    lowest_qual = sys.maxint
                    index = -1 # Position of record in group with lowest quality
                    for i,record in enumerate(group):
                        if record.FILTER:
                            if 'HETERO' in record.FILTER:
                                filter = record.FILTER
                                HETERO = True
                            if record.QUAL < lowest_qual:
                                index = i
                                lowest_qual = record.QUAL

                    if HETERO == False:
                           filter = group[0].FILTER
                    entry = group[index]
                    chrom = entry.CHROM
                    pos = new_pos
                    id = entry.ID
                    qual = entry.QUAL
                    info = entry.INFO
                    format = entry.FORMAT
                    samples = entry.samples
                    vcf_entry = vcf.model._Record(chrom, pos, id, new_refs, [new_alts], qual, filter, info, format,
                                                  samples)
                    merged_variants.append(vcf_entry)
                    continue # Next group of consecutive variants

                elif compare(unknown_cases, ['deletion', 'snp', 'insertion']) == True:
                    for item in group:
                        merged_variants.append(item)
                    continue

            elif len(group) == 2:
                # Remain in the same group and leave processing for next set of statements
                pass
            elif len(group) == 1: # Ex: ['snp','snp','snp'] --> ['snp']
                merged_variants.append(group[0])
                continue # Iterate to next group of consecutive variants
            elif len(group) == 4:
                combined_refs, combined_alts = "",""
                for index, record in enumerate(group): #Simple concatenation
                    combined_refs += record.REF
                    combined_alts += str(record.ALT[0])

                HETERO = False
                filter = ""
                lowest_qual = sys.maxint
                index = -1 # Position of record in group with lowest quality
                for i,record in enumerate(group):
                    if record.FILTER:
                        if 'HETERO' in record.FILTER:
                            filter = record.FILTER
                            HETERO = True
                        if record.QUAL < lowest_qual:
                            index = i
                            lowest_qual = record.QUAL

                if HETERO == False:
                       filter = group[0].FILTER
                entry = group[index]
                chrom = entry.CHROM
                pos = group[0].POS # Assign position number of first variant in group
                id = entry.ID
                qual = entry.QUAL
                info = entry.INFO
                format = entry.FORMAT
                samples = entry.samples
                vcf_entry = vcf.model._Record(chrom, pos, id, combined_refs, [combined_alts], qual, filter, info,
                                              format, samples)
                merged_variants.append(vcf_entry)
                continue
            else:
                logger.warn("unknown case: ", unknown_cases, unknown_refs_alts)

        if len(group) == 2:
            for index, cur_entry in enumerate(group):
                if index == 0: # We are on the first entry of the current consec group
                    previous_id = check_type(cur_entry)[0]
                    prev_entry = cur_entry

                else:
                    current_id = check_type(cur_entry)[0]
                    if previous_id == "insertion":
                        if current_id == "insertion":
                            for item in group:
                                merged_variants.append(item)
                            break
                        elif current_id == "deletion":
                            ref1 = populate_list(prev_entry.REF, prev_entry.POS)
                            ref2 = populate_list(cur_entry.REF, cur_entry.POS)
                            alt1 = populate_list(str(prev_entry.ALT[0]), prev_entry.POS)
                            alt2 = populate_list(str(cur_entry.ALT[0]), cur_entry.POS)

                            combined_ref = ref1 + ref2
                            combined_alt = alt1 + alt2
                            new_refs,new_alts= "",""
                            if len(combined_ref)==len(combined_alt):
                                first_occurence = -1
                                for i,j in zip(combined_ref,combined_alt):
                                    if i[0]!=j[0]:
                                        if first_occurence == -1:
                                            new_pos = i[1]
                                            first_occurence += 1
                                        new_refs += i[0]
                                        new_alts += j[0]

                                HETERO = False
                                lowest_qual = sys.maxint
                                index = -1 # Position of record in group with lowest quality
                                for i,record in enumerate(group):
                                    if record.FILTER:
                                        if 'HETERO' in record.FILTER:
                                            filter = record.FILTER
                                            HETERO = True
                                    if record.QUAL < lowest_qual:
                                        index = i
                                    lowest_qual = record.QUAL

                                if HETERO == False:
                                    filter = group[0].FILTER
                                entry = group[index]
                                chrom = entry.CHROM
                                pos = new_pos
                                id = entry.ID
                                qual = entry.QUAL
                                info = entry.INFO
                                format = entry.FORMAT
                                samples = entry.samples
                                vcf_entry = vcf.model._Record(chrom, new_pos, id, new_refs, [new_alts], qual, filter,
                                                              info,format,samples)
                                merged_variants.append(vcf_entry)
                        elif current_id == "snp":
                            merged_variants.append(prev_entry)
                            merged_variants.append(cur_entry)
                            break
                    elif previous_id == "deletion":
                        if current_id == "insertion":
                            ref1 = populate_list(prev_entry.REF, prev_entry.POS)
                            ref2 = populate_list(cur_entry.REF, cur_entry.POS)
                            alt1 = populate_list(str(prev_entry.ALT[0]), prev_entry.POS)
                            alt2 = populate_list(str(cur_entry.ALT[0]), cur_entry.POS)
                            combined_ref = sorted(set(ref1+ref2), key=itemgetter(1)) # populated list
                            combined_alt = alt1 # populated list

                            for index, entry in enumerate(alt2):
                                if index != 0:
                                    combined_alt.append(entry)

                            combined_alt = sorted(combined_alt, key=itemgetter(1))
                            if len(combined_ref)==len(combined_alt):
                                new_ref, new_alt= "", ""

                                first_occurence = -1
                                for i,j in zip(combined_ref,combined_alt):
                                    if i[0] != j[0]:
                                        if first_occurence == -1:
                                            new_pos = i[1]
                                            first_occurence += 1
                                        new_ref += i[0]
                                        new_alt += j[0]

                                HETERO = False
                                filter = ''
                                lowest_qual = sys.maxint
                                index = -1 # Position of record in group with lowest quality
                                for i,record in enumerate(group):
                                    if record.FILTER:
                                        if 'HETERO' in record.FILTER:
                                            filter = record.FILTER
                                            HETERO = True
                                    if record.QUAL < lowest_qual:
                                        index = i
                                        lowest_qual = record.QUAL
                                if HETERO == False:
                                    filter = group[0].FILTER
                                entry = group[index]
                                chrom = entry.CHROM
                                pos = new_pos
                                id = entry.ID
                                qual = entry.QUAL
                                info = entry.INFO
                                format = entry.FORMAT
                                samples = entry.samples
                                vcf_entry = vcf.model._Record(chrom, new_pos, id, new_ref,[new_alt], qual, filter, info,
                                                              format, samples)
                                merged_variants.append(vcf_entry)
                                continue

                            # Lengths are not equal: del/ins can't be merged
                            else:
                                merged_variants.append(prev_entry)
                                merged_variants.append(cur_entry)
                                break

                        elif current_id == "deletion":
                            ref1_list = populate_list(prev_entry.REF, prev_entry.POS)
                            ref2_list = populate_list(cur_entry.REF, cur_entry.POS)
                            set_combined_refs = set(ref1_list + ref2_list)
                            sorted_combined_refs = sorted(set_combined_refs, key=itemgetter(1)) # sort unique refs by position number (second item in tuple)
                            combined_refs = ""
                            for unique_union in sorted_combined_refs:
                                combined_refs += unique_union[0]
                            combined_alts = str(prev_entry.ALT[0])
                            HETERO = False
                            filter = ''
                            lowest_qual = sys.maxint
                            index = -1 # Position of record in group with lowest quality
                            for i,record in enumerate(group):
                                if record.FILTER:
                                    if 'HETERO' in record.FILTER:
                                        filter = record.FILTER
                                        HETERO = True
                                if record.QUAL < lowest_qual:
                                    index = i
                                    lowest_qual = record.QUAL
                            if HETERO == False:
                                filter = group[0].FILTER
                            entry = group[index]
                            chrom = entry.CHROM
                            pos = prev_entry.POS
                            id = entry.ID
                            qual = entry.QUAL
                            info = entry.INFO
                            format = entry.FORMAT
                            samples = entry.samples
                            vcf_entry = vcf.model._Record(chrom, pos, id, combined_refs, combined_alts, qual, filter,
                                                          info, format, samples)
                            merged_variants.append(vcf_entry)
                            continue

                        elif current_id == "snp":
                            for index,item in enumerate(group):
                                if index != 0:
                                    merged_variants.append(item)
                            break  # Break out of for loop

                    elif previous_id == "snp":
                        if (current_id == "insertion") or (current_id == "deletion"):
                            for item in group:
                                merged_variants.append(item)
                            break

                        elif current_id == "snp":
                            if cur_entry.POS == prev_entry.POS + 1:
                                if len(prev_entry.ALT) == 1:
                                    prev_alt = prev_entry.ALT[0]
                                else:
                                    prev_alt = str(prev_entry.ALT[0])
                                if len(cur_entry.ALT) == 1:
                                    cur_alt = cur_entry.ALT[0]
                                else:
                                    cur_alt = str(cur_entry.ALT[0])
                                combined_ref = prev_entry.REF + cur_entry.REF
                                combined_alt = str(prev_alt) + str(cur_alt)

                                HETERO = False
                                lowest_qual = sys.maxint
                                index = -1 # Position of record in group with lowest quality

                                for i,record in enumerate(group):
                                    if record.FILTER:
                                        if 'HETERO' in record.FILTER:
                                            filter = record.FILTER
                                            HETERO = True
                                    if record.QUAL < lowest_qual:
                                        index = i
                                        lowest_qual = record.QUAL
                                    if HETERO == False:
                                        filter = group[0].FILTER
                                entry = group[index]
                                chrom = entry.CHROM
                                pos = group[0].POS # new position will be first entry in group
                                id = entry.ID
                                qual = entry.QUAL
                                info = entry.INFO
                                format = entry.FORMAT
                                samples = entry.samples
                                vcf_entry = vcf.model._Record(chrom, pos, id, combined_ref, [combined_alt], qual,
                                                              filter, info, format, samples)
                                merged_variants.append(vcf_entry)
                            elif cur_entry.POS != prev_entry.POS + 1:
                                for item in group:
                                    merged_variants.append(item)
                                break

                        previous_id = current_id
                        prev_entry = cur_entry
    return merged_variants

def get_top_variant(variant_coverage, qualities):
    """Select a top variant."""

    coverages = [_ for _ in variant_coverage.itervalues()]
    max_coverage = max(coverages)

    top_variants = [call for call, coverage in variant_coverage.iteritems()
                    if coverage == max_coverage]
    top_qualities = {call:quality for call, quality in qualities.iteritems()
                     if call in top_variants}

    if coverages.count(max_coverage) > 1:
        snps = [var for var in top_variants
                if is_snp(var)]
        single_indels = [var for var in top_variants
                         if is_single_indel(var)]
        multi_indels = [var for var in top_variants
                        if is_multibase_indel(var)]

        if snps:
            max_base_quality = max(top_qualities.values())
            top_snp = [snp for snp in snps
                       if qualities[snp]==max_base_quality]
        else:
            top_snp = None

        score=0
        if multi_indels:
            for indel in multi_indels:
                if(qualities[indel]>score):
                    top_indel = indel
                    score=qualities[indel]
        elif single_indels:
            for indel in single_indels:
                if(qualities[indel]>score):
                    top_indel = indel
                    score=qualities[indel]
        else:
            top_indel = None

        return top_indel if not top_snp else top_snp[0]
    else:
        return top_variants[0]


def eval_variant(variant, RSR, VSR, var_freq=None, multi_del=False):
    """
    Evaluate variants against a reference.

    Given the variant (for which we only really need the type),
    the number of supporting reads for the reference (RSR)
    and the variant (VSR), and the variant frequency,
    return the decision as a VCF filter string (either
    PASS, HETERO, LOW, or NO).
    """
    if not var_freq:
        var_freq = VSR/(RSR + VSR)

    if VSR < 4:
        if RSR < 4:
            return 'LOW'
        else:
            return 'NO'

    if is_multibase_indel(variant) or is_snp(variant):
        if RSR >= 4:
            if var_freq < (1/3):
                return 'NO'
            elif var_freq <= (2/3):
                return 'HETERO'
            else:
                return 'PASS'
        else:
            return 'PASS'

    elif is_single_indel(variant):
        if VSR < 6 and RSR < 4:
            return 'LOW'
        elif RSR > 0:
            if var_freq < (1/2) or VSR < 6:
                return 'NO'
            elif var_freq < (2/3):
                return 'HETERO'
            else:
                return 'PASS'
        else:
            return 'PASS'

