#!/usr/bin/env python2.7

__author__ = 'Kellie Kim'

import unittest
import sys
import os
import subprocess
sys.path.append('pbhoover/')
from PBHoover_Runner import target_region_run, multi_chunk_run, write_to_vcf_single
from pysam import VariantFile


def sanger_snp_dict(sanger_location, sanger_snp_file, input_isolate):
    """Builds a dictionary that contains the isolate and the corresponding mutations called by Sanger"""
    sanger_dict = {}
    with open(sanger_location + sanger_snp_file, 'r') as sanger_filehandle:
        for line in sanger_filehandle:
            column = line.rstrip('\n').split('\t')
            if 'Isolate_Id' in column[0]:
                continue
            isolate=column[0]
            gene = column[2]
            position = int(column[3])
            alt = column[-1]
            if isolate == input_isolate:
                if isolate not in sanger_dict.keys():
                    sanger_dict[isolate] = {}
                sanger_dict[isolate][position] = alt
    return sanger_dict

class TestClass(unittest.TestCase):
	# Static vars accessible by any instance of this class
	sanger_location = "pbhoover/tests/" #os.getcwd()
	sanger_snp_file = sanger_location + "sanger_snp_all.txt"

	# The steps in this method will be performed before each test is run
	def setUp(self):
		self.sanger_dictionary = None

	# Test 1: Same SNP called by Sanger and PBHoover
	def test_1_same_snp(self):
		devnull = open(os.devnull,'w')
		# Set-up arguments for pbhoover target_region_run
		WGS_REF, WGS_ALT, WGS_FILTER = None,None,None
		input_isolate = '4-0101' 
		input_h5_file = 'pbhoover/tests/4-0101_sorted_with_header.1917972-1917972.corrected.cmp.h5' 
		output_file = '4-0101.1917972-1917972.vcf'
		target_genome_position = 1917972
		target_region_run(False,input_h5_file,output_file,(target_genome_position-20,target_genome_position+20))	 			 	
		p1 = subprocess.Popen('bgzip -f 4-0101.1917972-1917972.vcf', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p1.wait() # Wait for process 1 to complete
		p2 = subprocess.Popen('tabix -f -p vcf 4-0101.1917972-1917972.vcf.gz', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p2.wait() # Wait until process 2 is completed before moving onto next step
		# Open .vcf.gz file created from partial run of PBHoover
		try:
			vcf_in = VariantFile("4-0101.1917972-1917972.vcf.gz")	
			for record in vcf_in.fetch():
				if record.pos == target_genome_position:
					WGS_REF = record.ref
					WGS_ALT = ''.join(record.alts)
					WGS_FILTER = record.filter.keys()[0] # filter object is a dictionary in pysam, need to return a list from keys method and select item at first index

		except:
			print "No variants called in vcf file"

		 # Set TestClass sanger dictionary attribute
		self.sanger_dictionary = sanger_snp_dict(self.sanger_location, self.sanger_snp_file, input_isolate)
		try:
			Sanger_SNP_call = self.sanger_dictionary[input_isolate][target_genome_position]
		except:
			Sanger_SNP_call = None
		self.assertTrue(((WGS_REF == 'A') and (WGS_ALT == 'G') and (Sanger_SNP_call == 'G')), msg=\
		"Failed Test 1: Same SNP not called by Sanger and PBHoover")
		print "Passed Test 1"

	# Test 2: WGS Low, Sanger SNP
	def test_2_wgs_low_sanger_snp(self):
		WGS_REF, WGS_ALT, WGS_FILTER = None,None,None
		input_isolate = "4-0077"
		input_h5_file = "pbhoover/tests/4-0077_sorted_with_header.1673425-1673425.corrected.cmp.h5" 
		output_file = "4-0077.1673425-1673425.vcf"
		target_genome_position = 1673425
		target_region_run(False, input_h5_file, output_file, (target_genome_position-20,target_genome_position+20))
		p1 = subprocess.Popen('bgzip -f 4-0077.1673425-1673425.vcf', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p1.wait()
		p2 = subprocess.Popen('tabix -f -p vcf 4-0077.1673425-1673425.vcf.gz', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p2.wait() 
		try:
			vcf_in = VariantFile("4-0077.1673425-1673425.vcf.gz")	
			for record in vcf_in.fetch():
				if record.pos == target_genome_position:
					WGS_REF = record.ref
					WGS_ALT = ''.join(record.alts)
					WGS_FILTER = record.filter.keys()[0] # filter object is a dictionary in pysam, need to return a list from keys method and select item at first index
		except:
			pass

		self.sanger_dictionary = sanger_snp_dict(self.sanger_location, self.sanger_snp_file, input_isolate)
		self.assertTrue(((WGS_FILTER == None) and (self.sanger_dictionary[input_isolate][target_genome_position] == 'T')), msg=\
		"Failed Test 2: WGS filter not low or Sanger did not call SNP")
		print "Passed Test 2"

	# Test 3: WGS HETERO Filter, Sanger Same Variant
	def test_3_wgs_hetero_sanger_same(self):
		WGS_REF, WGS_ALT, WGS_FILTER = None,None,None
		input_isolate = "4-0023" 
		input_h5_file = "pbhoover/tests/4-0023_sorted_with_header.2155167-2155167.corrected.cmp.h5"
		output_file = '4-0023.2155167-2155167.vcf'
		target_genome_position = 2155167
		target_region_run(False,input_h5_file,output_file,(target_genome_position-20,target_genome_position+20))
		p1 = subprocess.Popen('bgzip -f 4-0023.2155167-2155167.vcf', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE)
		p1.wait()
		p2 = subprocess.Popen('tabix -f -p vcf 4-0023.2155167-2155167.vcf.gz', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE)
		p2.wait()
		try:
			vcf_in = VariantFile("4-0023.2155167-2155167.vcf.gz")	
			for record in vcf_in.fetch():
				if (record.pos == target_genome_position) and (record.filter.keys()[0] == "HETERO"):
					WGS_REF = record.ref
					WGS_ALT = ''.join(record.alts)
					WGS_FILTER = record.filter.keys()[0] # filter object is a dictionary in pysam, need to return a list from keys method and select item at first index

		except:
			print "No variants called in vcf file"

		 # Set TestClass sanger dictionary attribute
		self.sanger_dictionary = sanger_snp_dict(self.sanger_location, self.sanger_snp_file, input_isolate) 
		subprocess.Popen('rm 4-0023.2155167-2155167.vcf.gz.tbi', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		self.assertTrue(((WGS_FILTER == "HETERO") and (WGS_ALT == self.sanger_dictionary[input_isolate][target_genome_position])), msg=\
		"Failed Test 3: WGS filter not HETERO or Sanger did not call the same variant")
		print "Passed Test 3"


	# Test 4: WGS Low, Sanger Reference
	def test_4_wgs_low_sanger_reference(self):
		WGS_REF, WGS_ALT, WGS_FILTER = None,None,None
		input_isolate = "1-0033"
		input_h5_file = "pbhoover/tests/1-0033_sorted_with_header.2155114-2155114.corrected.cmp.h5"
		output_file = '1-0033.2155114-2155114.vcf'
		target_genome_position = 2155114
		target_region_run(False,input_h5_file,output_file,(target_genome_position-20,target_genome_position+20))
		p1 = subprocess.Popen('bgzip -f 1-0033.2155114-2155114.vcf', \
			shell=True, 
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p1.wait()
		p2 = subprocess.Popen('tabix -f -p vcf 1-0033.2155114-2155114.vcf.gz', \
			shell=True, 
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p2.wait()
		try:
			vcf_in = VariantFile("1-0033.2155114-2155114.vcf.gz")	
			for record in vcf_in.fetch():
				if record.pos == target_genome_position:
					WGS_REF = record.ref
					WGS_ALT = ''.join(record.alts)
					WGS_FILTER = record.filter.keys()[0] # filter object is a dictionary in pysam, need to return a list from keys method and select item at first index
		except:
			pass

		self.sanger_dictionary = sanger_snp_dict(self.sanger_location, self.sanger_snp_file, input_isolate) 
		self.assertTrue(((WGS_FILTER == None) and (target_genome_position not in self.sanger_dictionary[input_isolate].keys())), msg=\
		"Failed Test 4: WGS filter not low or Sanger did not call reference")
		print "Passed Test 4"

	# Test 5: WGS Variant, Sanger Reference
	def test_5_wgs_variant_sanger_reference(self):
		WGS_REF, WGS_ALT, WGS_FILTER = None,None,None
		input_isolate = "4-0097"
		input_h5_file = "pbhoover/tests/4-0097_sorted_with_header.1673470-1673470.corrected.cmp.h5" 
		output_file = '4-0097.1673470-1673470.vcf'
		target_genome_position = 1673470
		target_region_run(False,input_h5_file,output_file,(target_genome_position-20,target_genome_position+20))
		p1 = subprocess.Popen('bgzip -f 4-0097.1673470-1673470.vcf', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p1.wait()
		p2 = subprocess.Popen('tabix -f -p vcf 4-0097.1673470-1673470.vcf.gz', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p2.wait()
		try:
			vcf_in = VariantFile("4-0097.1673470-1673470.vcf.gz")	
			for record in vcf_in.fetch():
				if record.pos == target_genome_position:
					WGS_REF = record.ref
					WGS_ALT = ''.join(record.alts)
					WGS_FILTER = record.filter.keys()[0] # filter object is a dictionary in pysam, need to return a list from keys method and select item at first index
		except:
			print "No variants called in vcf file"

		self.sanger_dictionary = sanger_snp_dict(self.sanger_location, self.sanger_snp_file, input_isolate) 
		self.assertTrue(((WGS_REF == "CCC") and (WGS_ALT == 'C') and (target_genome_position not in self.sanger_dictionary[input_isolate].keys())), msg=\
		"Failed Test 5: WGS did not call variant, or Sanger did not call reference")
		print "Passed Test 5"
	
	# Test 6: WGS HETERO Filter, Sanger Reference	
	def test_6_wgs_hetero_sanger_reference(self):
		WGS_REF, WGS_ALT, WGS_FILTER = None,None,None
		input_isolate = "4-0103"
		input_h5_file = "pbhoover/tests/4-0103_sorted_with_header.7582-7582.corrected.cmp.h5"
		output_file = '4-0103.7582-7582.vcf'
		target_genome_position = 7582
		target_region_run(False,input_h5_file,output_file,(target_genome_position-20,target_genome_position+20))
		p1 = subprocess.Popen('bgzip -f 4-0103.7582-7582.vcf', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p1.wait()
		p2 = subprocess.Popen('tabix -f -p vcf 4-0103.7582-7582.vcf.gz', \
			shell=True, \
			stdin=subprocess.PIPE, \
			stdout=subprocess.PIPE, \
			stderr=subprocess.STDOUT)
		p2.wait()
		try:
			vcf_in = VariantFile("4-0103.7582-7582.vcf.gz")	
			for record in vcf_in.fetch():
				if (record.pos == target_genome_position) and (record.filter.keys()[0] == "HETERO"):
					WGS_REF = record.ref
					WGS_ALT = ''.join(record.alts)
					WGS_FILTER = record.filter.keys()[0] # filter object is a dictionary in pysam, need to return a list from keys method and select item at first index
		except:
			print "No variants called in vcf file"

		self.sanger_dictionary = sanger_snp_dict(self.sanger_location, self.sanger_snp_file, input_isolate) 
		self.assertTrue(((WGS_FILTER == "HETERO") and (target_genome_position not in self.sanger_dictionary[input_isolate].keys())), msg=\
		"Failed Test 6: WGS filter not HETERO or Sanger did not call reference")
		print "Passed Test 6"

if __name__ == '__main__':
	unittest.main() 
